@@ -2,7 +2,7 @@
  * This program source code file is part of KiCad, a free EDA CAD application.
  *
  * Copyright (C) 2021 Andrew Lutsenko, anlutsenko at gmail dot com
- * Copyright (C) 1992-2021 KiCad Developers, see AUTHORS.txt for contributors.
+ * Copyright The KiCad Developers, see AUTHORS.txt for contributors.
  *
  * This program is free software: you can redistribute it and/or modify it
  * under the terms of the GNU General Public License as published by the
@@ -20,15 +20,15 @@
 
 #include "panel_packages_view.h"
 #include <grid_tricks.h>
-#include <kicad_settings.h>
+#include <widgets/html_window.h>
 #include <pgm_base.h>
-#include <settings/settings_manager.h>
+#include <kiplatform/ui.h>
 #include <settings/common_settings.h>
-#include <widgets/wx_splitter_window.h>
-#include <widgets/wx_panel.h>
+#include <settings/kicad_settings.h>
+#include <settings/settings_manager.h>
 #include <string_utils.h>
-#include <html_window.h>
-#include <wildcards_and_files_ext.h>
+#include <widgets/wx_panel.h>
+#include <widgets/wx_splitter_window.h>
 
 #include <cmath>
 #include <fstream>
@@ -41,17 +41,21 @@
 
 
 std::unordered_map<PCM_PACKAGE_VERSION_STATUS, wxString> PANEL_PACKAGES_VIEW::STATUS_ENUM_TO_STR = {
-    { PVS_INVALID, wxT( "invalid" ) },
-    { PVS_STABLE, wxT( "stable" ) },
-    { PVS_TESTING, wxT( "testing" ) },
-    { PVS_DEVELOPMENT, wxT( "development" ) },
-    { PVS_DEPRECATED, wxT( "deprecated" ) }
+    { PVS_INVALID, wxS( "invalid" ) },
+    { PVS_STABLE, wxS( "stable" ) },
+    { PVS_TESTING, wxS( "testing" ) },
+    { PVS_DEVELOPMENT, wxS( "development" ) },
+    { PVS_DEPRECATED, wxS( "deprecated" ) }
 };
 
 
 PANEL_PACKAGES_VIEW::PANEL_PACKAGES_VIEW( wxWindow*                               parent,
-                                          std::shared_ptr<PLUGIN_CONTENT_MANAGER> aPcm ) :
+                                          std::shared_ptr<PLUGIN_CONTENT_MANAGER> aPcm,
+                                          const ActionCallback&                   aActionCallback,
+                                          const PinCallback&                      aPinCallback ) :
         PANEL_PACKAGES_VIEW_BASE( parent ),
+        m_actionCallback( aActionCallback ),
+        m_pinCallback( aPinCallback ),
         m_pcm( aPcm )
 {
     // Replace wxFormBuilder's sash initializer with one which will respect m_initialSashPos.
@@ -61,7 +65,7 @@ PANEL_PACKAGES_VIEW::PANEL_PACKAGES_VIEW( wxWindow*
     m_splitter1->Connect( wxEVT_IDLE, wxIdleEventHandler( PANEL_PACKAGES_VIEW::SetSashOnIdle ),
                           NULL, this );
 
-    m_splitter1->SetPaneMinimums( 350, 450 );
+    m_splitter1->SetPaneMinimums( FromDIP( 350 ), FromDIP( 450 ) );
 
 #ifdef __WXGTK__
     // wxSearchCtrl vertical height is not calculated correctly on some GTK setups
@@ -83,8 +87,7 @@ PANEL_PACKAGES_VIEW::PANEL_PACKAGES_VIEW( wxWindow*
 
         // Set the minimal width to the column label size.
         m_gridVersions->SetColMinimalWidth( col, headingWidth );
-        m_gridVersions->SetColSize( col, m_gridVersions->GetVisibleWidth( col, true, true,
-                                                                          false ) );
+        m_gridVersions->SetColSize( col, m_gridVersions->GetVisibleWidth( col ) );
     }
 
     // Most likely should be changed to wxGridSelectNone once WxWidgets>=3.1.5 is mandatory.
@@ -102,6 +105,9 @@ PANEL_PACKAGES_VIEW::PANEL_PACKAGES_VIEW( wxWindow*
 
 PANEL_PACKAGES_VIEW::~PANEL_PACKAGES_VIEW()
 {
+    m_splitter1->Disconnect( wxEVT_IDLE, wxIdleEventHandler( PANEL_PACKAGES_VIEW::SetSashOnIdle ),
+                             NULL, this );
+
     COMMON_SETTINGS* cfg = Pgm().GetCommonSettings();
     cfg->m_PackageManager.sash_pos = m_splitter1->GetSashPosition();
 
@@ -114,6 +120,7 @@ void PANEL_PACKAGES_VIEW::ClearData()
     unsetPackageDetails();
 
     m_currentSelected = nullptr;
+    m_updateablePackages.clear();
     m_packagePanels.clear();
     m_packageInitialOrder.clear();
     m_packageListWindow->GetSizer()->Clear( true ); // Delete panels
@@ -122,20 +129,17 @@ void PANEL_PACKAGES_VIEW::ClearData()
 }
 
 
-void PANEL_PACKAGES_VIEW::SetData( const std::vector<PACKAGE_VIEW_DATA>& aPackageData,
-                                   ActionCallback                        aCallback )
+void PANEL_PACKAGES_VIEW::SetData( const std::vector<PACKAGE_VIEW_DATA>& aPackageData )
 {
-    m_actionCallback = aCallback;
-
     ClearData();
 
     for( const PACKAGE_VIEW_DATA& data : aPackageData )
     {
-        PANEL_PACKAGE* package_panel =
-                new PANEL_PACKAGE( m_packageListWindow, m_actionCallback, data );
+        PANEL_PACKAGE* package_panel = new PANEL_PACKAGE( m_packageListWindow, m_actionCallback,
+                                                          m_pinCallback, data );
 
         package_panel->SetSelectCallback(
-                [package_panel, this]()
+                [package_panel, this] ()
                 {
                     if( m_currentSelected && m_currentSelected != package_panel )
                         m_currentSelected->SetSelected( false );
@@ -149,9 +153,13 @@ void PANEL_PACKAGES_VIEW::SetData( const std::vector<PACKAGE_VIEW_DATA>& aPackag
 
         m_packagePanels.insert( { data.package.identifier, package_panel } );
         m_packageInitialOrder.push_back( data.package.identifier );
+
+        if( data.state == PPS_UPDATE_AVAILABLE && !data.pinned )
+            m_updateablePackages.insert( data.package.identifier );
     }
 
     updatePackageList();
+    updateCommonState();
 }
 
 
@@ -177,19 +185,23 @@ void PANEL_PACKAGES_VIEW::setPackageDetails( const PACKAGE_VIEW_DATA& aPackageDa
 
                     if( inURL )
                     {
-                        if( c == ' ' )
+                        if( c == ' ' || c == '\n')
                         {
                             result += wxString::Format( wxT( "<a href='%s'>%s</a>" ), url, url );
                             inURL = false;
 
-                            result += c;
+                            if( c == '\n' )
+                                result += wxT( "</p><p>" );
+                            else
+                                result += c;
                         }
                         else
                         {
                             url += c;
                         }
                     }
-                    else if( text.Mid( i, 5 ) == wxT( "http:" ) || text.Mid( i, 6 ) == wxT( "https:" ) )
+                    else if( text.Mid( i, 5 ) == wxT( "http:" )
+                                || text.Mid( i, 6 ) == wxT( "https:" ) )
                     {
                         url = c;
                         inURL = true;
@@ -204,6 +216,9 @@ void PANEL_PACKAGES_VIEW::setPackageDetails( const PACKAGE_VIEW_DATA& aPackageDa
                     }
                 }
 
+                if( inURL )
+                    result += wxString::Format( wxT( "<a href='%s'>%s</a>" ), url, url );
+
                 return result;
             };
 
@@ -222,7 +237,7 @@ void PANEL_PACKAGES_VIEW::setPackageDetails( const PACKAGE_VIEW_DATA& aPackageDa
         for( const std::string& tag : package.tags )
         {
             if( !tags_str.IsEmpty() )
-                tags_str += wxT( ", " );
+                tags_str += ", ";
 
             tags_str += tag;
         }
@@ -250,7 +265,11 @@ void PANEL_PACKAGES_VIEW::setPackageDetails( const PACKAGE_VIEW_DATA& aPackageDa
                 details << wxT( "<li>" ) + type + wxT( ": " ) + contact.name + wxT( "<ul>" );
 
                 for( const std::pair<const std::string, wxString>& entry : contact.contact )
-                    details << wxT( "<li>" ) + entry.first + wxT( ": " ) + format_entry( entry ) + wxT( "</li>" );
+                {
+                    details << wxT( "<li>" );
+                    details << entry.first << wxT( ": " ) + format_entry( entry );
+                    details << wxT( "</li>" );
+                }
 
                 details << wxT( "</ul>" );
             };
@@ -258,14 +277,18 @@ void PANEL_PACKAGES_VIEW::setPackageDetails( const PACKAGE_VIEW_DATA& aPackageDa
     write_contact( _( "Author" ), package.author );
 
     if( package.maintainer )
-        write_contact( _( "Maintainer" ), package.maintainer.get() );
+        write_contact( _( "Maintainer" ), *package.maintainer );
 
     if( package.resources.size() > 0 )
     {
         details << wxT( "<li>" ) + _( "Resources" ) + wxT( "<ul>" );
 
         for( const std::pair<const std::string, wxString>& entry : package.resources )
-            details << wxT( "<li>" ) + entry.first + wxS( ": " ) + format_entry( entry ) + wxT( "</li>" );
+        {
+            details << wxT( "<li>" );
+            details << entry.first << wxT( ": " );
+            details << format_entry( entry ) + wxT( "</li>" );
+        }
 
         details << wxT( "</ul>" );
     }
@@ -286,7 +309,7 @@ void PANEL_PACKAGES_VIEW::setPackageDetails( const PACKAGE_VIEW_DATA& aPackageDa
     int      row = 0;
     wxString current_version;
 
-    if( aPackageData.state == PPS_INSTALLED )
+    if( aPackageData.state == PPS_INSTALLED || aPackageData.state == PPS_UPDATE_AVAILABLE )
         current_version = m_pcm->GetInstalledPackageVersion( package.identifier );
 
     wxFont bold_font = m_gridVersions->GetDefaultCellFont().Bold();
@@ -321,13 +344,16 @@ void PANEL_PACKAGES_VIEW::setPackageDetails( const PACKAGE_VIEW_DATA& aPackageDa
     for( int col = 0; col < m_gridVersions->GetNumberCols(); col++ )
     {
         // Set the width to see the full contents
-        m_gridVersions->SetColSize( col, m_gridVersions->GetVisibleWidth( col, true, true,
-                                                                          false ) );
+        m_gridVersions->SetColSize( col, m_gridVersions->GetVisibleWidth( col ) );
     }
 
+    // Autoselect preferred or installed version
     if( m_gridVersions->GetNumberRows() >= 1 )
     {
-        wxString version = m_currentSelected->GetPreferredVersion();
+        wxString version = m_currentSelected->GetPackageData().current_version;
+
+        if( version.IsEmpty() )
+            version = m_currentSelected->GetPreferredVersion();
 
         if( !version.IsEmpty() )
         {
@@ -335,8 +361,8 @@ void PANEL_PACKAGES_VIEW::setPackageDetails( const PACKAGE_VIEW_DATA& aPackageDa
             {
                 if( m_gridVersions->GetCellValue( i, COL_VERSION ) == version )
                 {
-                    std::cout << "auto select row: " << i << std::endl;
                     m_gridVersions->SelectRow( i );
+                    m_gridVersions->SetGridCursor( i, COL_VERSION );
                     break;
                 }
             }
@@ -380,20 +406,20 @@ void PANEL_PACKAGES_VIEW::unsetPackageDetails()
 }
 
 
-wxString PANEL_PACKAGES_VIEW::toHumanReadableSize( const boost::optional<uint64_t> size ) const
+wxString PANEL_PACKAGES_VIEW::toHumanReadableSize( const std::optional<uint64_t> size ) const
 {
     if( !size )
         return wxT( "-" );
 
-    uint64_t b = size.get();
+    uint64_t b = *size;
 
     if( b >= 1024 * 1024 )
-        return wxString::Format( wxT( "%.1f Mb" ), b / 1024.0 / 1024.0 );
+        return wxString::Format( wxT( "%.1f MB" ), b / 1000.0 / 1000.0 );
 
     if( b >= 1024 )
-        return wxString::Format( wxT( "%lld Kb" ), b / 1024 );
+        return wxString::Format( wxT( "%lld kB" ), b / 1000 );
 
-    return wxString::Format( wxT( "%lld b" ), b );
+    return wxString::Format( wxT( "%lld B" ), b );
 }
 
 
@@ -406,34 +432,50 @@ bool PANEL_PACKAGES_VIEW::canDownload() const
 }
 
 
-bool PANEL_PACKAGES_VIEW::canInstall() const
+bool PANEL_PACKAGES_VIEW::canRunAction() const
 {
     if( !m_currentSelected )
         return false;
 
     const PACKAGE_VIEW_DATA& packageData = m_currentSelected->GetPackageData();
 
-    if( packageData.state != PPS_AVAILABLE && packageData.state != PPS_UNAVAILABLE )
-        return false;
+    switch( packageData.state )
+    {
+    case PPS_PENDING_INSTALL:
+    case PPS_PENDING_UNINSTALL:
+    case PPS_PENDING_UPDATE: return false;
+    default: break;
+    }
 
     return m_gridVersions->GetNumberRows() == 1 || m_gridVersions->GetSelectedRows().size() == 1;
 }
 
 
 void PANEL_PACKAGES_VIEW::SetPackageState( const wxString&         aPackageId,
-                                           const PCM_PACKAGE_STATE aState ) const
+                                           const PCM_PACKAGE_STATE aState, const bool aPinned )
 {
     auto it = m_packagePanels.find( aPackageId );
 
     if( it != m_packagePanels.end() )
     {
-        it->second->SetState( aState );
+        it->second->SetState( aState, aPinned );
 
         if( m_currentSelected && m_currentSelected == it->second )
         {
             wxMouseEvent dummy;
             m_currentSelected->OnClick( dummy );
         }
+
+        if( aState == PPS_UPDATE_AVAILABLE && !aPinned )
+        {
+            m_updateablePackages.insert( aPackageId );
+        }
+        else
+        {
+            m_updateablePackages.erase( aPackageId );
+        }
+
+        updateCommonState();
     }
 }
 
@@ -469,34 +511,36 @@ void PANEL_PACKAGES_VIEW::OnDownloadVersionClicked( wxCommandEvent& event )
                                     return ver.version == version;
                                 } );
 
-    wxASSERT_MSG( ver_it != package.versions.end(), wxT( "Could not find package version" ) );
+    wxASSERT_MSG( ver_it != package.versions.end(), "Could not find package version" );
 
     if( !ver_it->download_url )
     {
         wxMessageBox( _( "Package download url is not specified" ),
-                      _( "Error downloading package" ), wxICON_INFORMATION | wxOK, this );
+                      _( "Error downloading package" ), wxICON_INFORMATION | wxOK,
+                      wxGetTopLevelParent( this ) );
         return;
     }
 
-    const wxString& url = ver_it->download_url.get();
+    const wxString& url = *ver_it->download_url;
 
     SETTINGS_MANAGER& mgr = Pgm().GetSettingsManager();
-    KICAD_SETTINGS*   app_settings = mgr.GetAppSettings<KICAD_SETTINGS>();
+    KICAD_SETTINGS*   cfg = mgr.GetAppSettings<KICAD_SETTINGS>( "kicad" );
 
-    wxFileDialog dialog( this, _( "Save package" ), app_settings->m_PcmLastDownloadDir,
+    wxWindow* topLevelParent = wxGetTopLevelParent( this );
+    wxFileDialog dialog( topLevelParent, _( "Save Package" ), cfg->m_PcmLastDownloadDir,
                          wxString::Format( wxT( "%s_v%s.zip" ), package.identifier, version ),
-                         ZipFileWildcard(), wxFD_SAVE | wxFD_OVERWRITE_PROMPT );
+                         wxT( "ZIP files (*.zip)|*.zip" ), wxFD_SAVE | wxFD_OVERWRITE_PROMPT );
 
     if( dialog.ShowModal() == wxID_CANCEL )
         return;
 
     wxString path = dialog.GetPath();
-    app_settings->m_PcmLastDownloadDir = wxPathOnly( path );
+    cfg->m_PcmLastDownloadDir = wxPathOnly( path );
 
     std::ofstream output( path.ToUTF8(), std::ios_base::binary );
 
-    std::unique_ptr<WX_PROGRESS_REPORTER> reporter(
-            new WX_PROGRESS_REPORTER( this, _( "Downloading package" ), 1 ) );
+    std::unique_ptr<WX_PROGRESS_REPORTER> reporter =
+            std::make_unique<WX_PROGRESS_REPORTER>( this, _( "Downloading package" ), 1 );
 
     bool success = m_pcm->DownloadToStream( url, &output, reporter.get(), 0 );
 
@@ -508,7 +552,7 @@ void PANEL_PACKAGES_VIEW::OnDownloadVersionClicked( wxCommandEvent& event )
         {
             std::ifstream stream( path.ToUTF8(), std::ios_base::binary );
 
-            bool matches = m_pcm->VerifyHash( stream, ver_it->download_sha256.get() );
+            bool matches = m_pcm->VerifyHash( stream, *ver_it->download_sha256 );
 
             stream.close();
 
@@ -516,7 +560,8 @@ void PANEL_PACKAGES_VIEW::OnDownloadVersionClicked( wxCommandEvent& event )
                 && wxMessageBox(
                            _( "Integrity of the downloaded package could not be verified, hash "
                               "does not match. Are you sure you want to keep this file?" ),
-                           _( "Keep downloaded file" ), wxICON_EXCLAMATION | wxYES_NO, this )
+                           _( "Keep downloaded file" ), wxICON_EXCLAMATION | wxYES_NO,
+                           wxGetTopLevelParent( this ) )
                            == wxNO )
             {
                 wxRemoveFile( path );
@@ -531,14 +576,23 @@ void PANEL_PACKAGES_VIEW::OnDownloadVersionClicked( wxCommandEvent& event )
 }
 
 
-void PANEL_PACKAGES_VIEW::OnInstallVersionClicked( wxCommandEvent& event )
+void PANEL_PACKAGES_VIEW::OnVersionActionClicked( wxCommandEvent& event )
 {
-    if( !canInstall() )
+    if( !canRunAction() )
     {
         wxBell();
         return;
     }
 
+    PCM_PACKAGE_ACTION action = getAction();
+
+    if( action == PPA_UNINSTALL )
+    {
+        m_actionCallback( m_currentSelected->GetPackageData(), PPA_UNINSTALL,
+                          m_currentSelected->GetPackageData().current_version );
+        return;
+    }
+
     if( m_gridVersions->GetNumberRows() == 1 )
         m_gridVersions->SelectRow( 0 );
 
@@ -553,18 +607,19 @@ void PANEL_PACKAGES_VIEW::OnInstallVersionClicked( wxCommandEvent& event )
                                     return ver.version == version;
                                 } );
 
-    wxASSERT_MSG( ver_it != package.versions.end(), wxT( "Could not find package version" ) );
+    wxCHECK_RET( ver_it != package.versions.end(), "Could not find package version" );
 
     if( !ver_it->compatible
-        && wxMessageBox( _( "This package version is incompatible with your kicad version or "
+        && wxMessageBox( _( "This package version is incompatible with your KiCad version or "
                             "platform. Are you sure you want to install it anyway?" ),
-                         _( "Install package" ), wxICON_EXCLAMATION | wxYES_NO, this )
+                         _( "Install package" ), wxICON_EXCLAMATION | wxYES_NO,
+                         wxGetTopLevelParent( this ) )
                    == wxNO )
     {
         return;
     }
 
-    m_actionCallback( m_currentSelected->GetPackageData(), PPA_INSTALL, version );
+    m_actionCallback( m_currentSelected->GetPackageData(), action, version );
 }
 
 
@@ -634,6 +689,12 @@ void PANEL_PACKAGES_VIEW::updatePackageList()
         {
             sizer->Add( panel, 0, wxEXPAND );
             panel->Show();
+
+            if( !m_currentSelected )
+            {
+                wxMouseEvent dummy;
+                panel->OnClick( dummy );
+            }
         }
         else
         {
@@ -650,14 +711,63 @@ void PANEL_PACKAGES_VIEW::updatePackageList()
 void PANEL_PACKAGES_VIEW::updateDetailsButtons()
 {
     m_buttonDownload->Enable( canDownload() );
-    m_buttonInstall->Enable( canInstall() );
+
+    if( canRunAction() )
+    {
+        m_buttonAction->Enable();
+
+        PCM_PACKAGE_ACTION action = getAction();
+
+        switch( action )
+        {
+        case PPA_INSTALL: m_buttonAction->SetLabel( _( "Install" ) ); break;
+        case PPA_UNINSTALL: m_buttonAction->SetLabel( _( "Uninstall" ) ); break;
+        case PPA_UPDATE: m_buttonAction->SetLabel( _( "Update" ) ); break;
+        }
+    }
+    else
+    {
+        m_buttonAction->Disable();
+        m_buttonAction->SetLabel( _( "Pending" ) );
+    }
+}
+
+
+PCM_PACKAGE_ACTION PANEL_PACKAGES_VIEW::getAction() const
+{
+    wxASSERT_MSG( m_gridVersions->GetNumberRows() == 1
+                          || m_gridVersions->GetSelectedRows().size() == 1,
+                  wxT( "getAction() called with ambiguous version selection" ) );
+
+    int selected_row = 0;
+
+    if( m_gridVersions->GetSelectedRows().size() == 1 )
+        selected_row = m_gridVersions->GetSelectedRows()[0];
+
+    wxString                 version = m_gridVersions->GetCellValue( selected_row, COL_VERSION );
+    const PACKAGE_VIEW_DATA& package = m_currentSelected->GetPackageData();
+
+    switch( package.state )
+    {
+    case PPS_AVAILABLE:
+    case PPS_UNAVAILABLE:
+        return PPA_INSTALL; // Only action for not installed package is to install it
+    case PPS_INSTALLED:
+    case PPS_UPDATE_AVAILABLE:
+        if( version == package.current_version )
+            return PPA_UNINSTALL;
+        else
+            return PPA_UPDATE;
+    default:
+        return PPA_INSTALL; // For pending states return value does not matter as button will be disabled
+    }
 }
 
 
 void PANEL_PACKAGES_VIEW::OnSizeInfoBox( wxSizeEvent& aEvent )
 {
-    wxSize infoSize = m_infoText->GetParent()->GetClientSize();
-    infoSize.x -= 8;
+    wxSize infoSize = KIPLATFORM::UI::GetUnobscuredSize( m_infoText->GetParent() );
+    infoSize.x -= 10;
     m_infoText->SetMinSize( infoSize );
     m_infoText->SetMaxSize( infoSize );
     m_infoText->SetSize( infoSize );
@@ -694,6 +804,34 @@ void PANEL_PACKAGES_VIEW::SetSashOnIdle( wxIdleEvent& aEvent )
 
     m_packageListWindow->FitInside();
 
-	m_splitter1->Disconnect( wxEVT_IDLE, wxIdleEventHandler( PANEL_PACKAGES_VIEW::SetSashOnIdle ),
+    m_splitter1->Disconnect( wxEVT_IDLE, wxIdleEventHandler( PANEL_PACKAGES_VIEW::SetSashOnIdle ),
                              NULL, this );
 }
+
+
+void PANEL_PACKAGES_VIEW::updateCommonState()
+{
+    m_buttonUpdateAll->Enable( m_updateablePackages.size() > 0 );
+}
+
+
+void PANEL_PACKAGES_VIEW::OnUpdateAllClicked( wxCommandEvent& event )
+{
+    // The map will be modified by the callback so we copy the list here
+    std::vector<wxString> packages;
+
+    std::copy( m_updateablePackages.begin(), m_updateablePackages.end(),
+               std::back_inserter( packages ) );
+
+    for( const wxString& pkg_id : packages )
+    {
+        auto it = m_packagePanels.find( pkg_id );
+
+        if( it != m_packagePanels.end() )
+        {
+            const PACKAGE_VIEW_DATA& data = it->second->GetPackageData();
+
+            m_actionCallback( data, PPA_UPDATE, data.update_version );
+        }
+    }
+}
