@@ -3,7 +3,7 @@
  *
  * Copyright (C) 2012-2015 Jean-Pierre Charras, jp.charras at wanadoo.fr
  * Copyright (C) 2008-2016 Wayne Stambaugh <stambaughw@gmail.com>
- * Copyright (C) 2004-2021 KiCad Developers, see AUTHORS.txt for contributors.
+ * Copyright The KiCad Developers, see AUTHORS.txt for contributors.
  *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License
@@ -34,12 +34,19 @@
 #include <footprint_viewer_frame.h>
 #include <fp_lib_table.h>
 #include <kiway.h>
+#include <kiway_express.h>
+#include <netlist_reader/pcb_netlist.h>
 #include <widgets/msgpanel.h>
+#include <widgets/wx_listbox.h>
+#include <widgets/wx_aui_utils.h>
+#include <gal/graphics_abstraction_layer.h>
 #include <pcb_draw_panel_gal.h>
 #include <pcb_painter.h>
 #include <pcbnew_id.h>
 #include <footprint_editor_settings.h>
 #include <pgm_base.h>
+#include <project_pcb.h>
+#include <project/project_file.h>
 #include <settings/settings_manager.h>
 #include <tool/action_toolbar.h>
 #include <tool/common_control.h>
@@ -56,20 +63,24 @@
 #include <tools/pcb_selection_tool.h>
 #include <tools/board_editor_control.h>
 #include <wildcards_and_files_ext.h>
-#include <wx/listbox.h>
+#include <lib_tree_model_adapter.h>
 #include <wx/srchctrl.h>
 #include <wx/tokenzr.h>
 #include <wx/choice.h>
+#include <wx/hyperlink.h>
+
+#include "invoke_pcb_dialog.h"
 
 using namespace std::placeholders;
 
 
-#define NEXT_PART       1
 #define NEW_PART        0
-#define PREVIOUS_PART   -1
+#define NEXT_PART       1
+#define PREVIOUS_PART   2
+#define RELOAD_PART     3
 
 
-BEGIN_EVENT_TABLE( FOOTPRINT_VIEWER_FRAME, EDA_DRAW_FRAME )
+BEGIN_EVENT_TABLE( FOOTPRINT_VIEWER_FRAME, PCB_BASE_FRAME )
     // Window events
     EVT_SIZE( FOOTPRINT_VIEWER_FRAME::OnSize )
     EVT_ACTIVATE( FOOTPRINT_VIEWER_FRAME::OnActivate )
@@ -92,34 +103,17 @@ BEGIN_EVENT_TABLE( FOOTPRINT_VIEWER_FRAME, EDA_DRAW_FRAME )
     // listbox events
     EVT_LISTBOX( ID_MODVIEW_LIB_LIST, FOOTPRINT_VIEWER_FRAME::ClickOnLibList )
     EVT_LISTBOX( ID_MODVIEW_FOOTPRINT_LIST, FOOTPRINT_VIEWER_FRAME::ClickOnFootprintList )
-    EVT_LISTBOX_DCLICK( ID_MODVIEW_FOOTPRINT_LIST, FOOTPRINT_VIEWER_FRAME::DClickOnFootprintList )
 
 END_EVENT_TABLE()
 
 
-/*
- * Note: FOOTPRINT_VIEWER_FRAME can be created in "modal mode", or as a usual frame.
- */
-#define PARENT_STYLE   ( KICAD_DEFAULT_DRAWFRAME_STYLE | wxFRAME_FLOAT_ON_PARENT )
-#define MODAL_STYLE    ( KICAD_DEFAULT_DRAWFRAME_STYLE | wxSTAY_ON_TOP )
-#define NONMODAL_STYLE ( KICAD_DEFAULT_DRAWFRAME_STYLE )
-
-
-FOOTPRINT_VIEWER_FRAME::FOOTPRINT_VIEWER_FRAME( KIWAY* aKiway, wxWindow* aParent,
-                                                FRAME_T aFrameType ) :
-    PCB_BASE_FRAME( aKiway, aParent, aFrameType, _( "Footprint Library Browser" ),
-            wxDefaultPosition, wxDefaultSize,
-            aFrameType == FRAME_FOOTPRINT_VIEWER_MODAL ? ( aParent ? PARENT_STYLE : MODAL_STYLE )
-                                                       : NONMODAL_STYLE,
-            aFrameType == FRAME_FOOTPRINT_VIEWER_MODAL ? FOOTPRINT_VIEWER_FRAME_NAME_MODAL
-                                                       : FOOTPRINT_VIEWER_FRAME_NAME )
+FOOTPRINT_VIEWER_FRAME::FOOTPRINT_VIEWER_FRAME( KIWAY* aKiway, wxWindow* aParent ) :
+    PCB_BASE_FRAME( aKiway, aParent, FRAME_FOOTPRINT_VIEWER, _( "Footprint Library Browser" ),
+                    wxDefaultPosition, wxDefaultSize, KICAD_DEFAULT_DRAWFRAME_STYLE,
+                    FOOTPRINT_VIEWER_FRAME_NAME ),
+   m_comp( LIB_ID(), wxEmptyString, wxEmptyString, KIID_PATH(), {} )
 {
-    wxASSERT( aFrameType == FRAME_FOOTPRINT_VIEWER_MODAL || aFrameType == FRAME_FOOTPRINT_VIEWER );
-
-    if( aFrameType == FRAME_FOOTPRINT_VIEWER_MODAL )
-        SetModal( true );
-
-    m_aboutTitle = _( "KiCad Footprint Library Viewer" );
+    m_aboutTitle = _HKI( "KiCad Footprint Library Browser" );
 
     // Force the items to always snap
     m_magneticItems.pads     = MAGNETIC_OPTIONS::CAPTURE_ALWAYS;
@@ -136,16 +130,19 @@ FOOTPRINT_VIEWER_FRAME::FOOTPRINT_VIEWER_FRAME( KIWAY* aKiway, wxWindow* aParent
     icon.CopyFromBitmap( KiBitmap( BITMAPS::icon_footprint_browser ) );
     SetIcon( icon );
 
+    m_libListWidth = 200;
+    m_fpListWidth = 300;
+
     wxPanel* libPanel = new wxPanel( this );
     wxSizer* libSizer = new wxBoxSizer( wxVERTICAL );
 
     m_libFilter = new wxSearchCtrl( libPanel, ID_MODVIEW_LIB_FILTER, wxEmptyString,
-                                  wxDefaultPosition, wxDefaultSize, wxTE_PROCESS_ENTER );
+                                    wxDefaultPosition, wxDefaultSize, wxTE_PROCESS_ENTER );
     m_libFilter->SetDescriptiveText( _( "Filter" ) );
     libSizer->Add( m_libFilter, 0, wxEXPAND, 5 );
 
-    m_libList = new wxListBox( libPanel, ID_MODVIEW_LIB_LIST, wxDefaultPosition, wxDefaultSize,
-                               0, nullptr, wxLB_HSCROLL | wxNO_BORDER );
+    m_libList = new WX_LISTBOX( libPanel, ID_MODVIEW_LIB_LIST, wxDefaultPosition, wxDefaultSize,
+                                0, nullptr, wxLB_HSCROLL | wxNO_BORDER );
     libSizer->Add( m_libList, 1, wxEXPAND, 5 );
 
     libPanel->SetSizer( libSizer );
@@ -155,7 +152,7 @@ FOOTPRINT_VIEWER_FRAME::FOOTPRINT_VIEWER_FRAME( KIWAY* aKiway, wxWindow* aParent
     wxSizer* fpSizer = new wxBoxSizer( wxVERTICAL );
 
     m_fpFilter = new wxSearchCtrl( fpPanel, ID_MODVIEW_FOOTPRINT_FILTER, wxEmptyString,
-                                 wxDefaultPosition, wxDefaultSize, wxTE_PROCESS_ENTER );
+                                   wxDefaultPosition, wxDefaultSize, wxTE_PROCESS_ENTER );
     m_fpFilter->SetDescriptiveText( _( "Filter" ) );
     m_fpFilter->SetToolTip(
             _( "Filter on footprint name, keywords, description and pad count.\n"
@@ -170,8 +167,12 @@ FOOTPRINT_VIEWER_FRAME::FOOTPRINT_VIEWER_FRAME( KIWAY* aKiway, wxWindow* aParent
     m_fpFilter->SetMinSize( wxSize( -1, GetTextExtent( wxT( "qb" ) ).y + 10 ) );
 #endif
 
-    m_fpList = new wxListBox( fpPanel, ID_MODVIEW_FOOTPRINT_LIST, wxDefaultPosition, wxDefaultSize,
-                              0, nullptr, wxLB_HSCROLL | wxNO_BORDER );
+    m_fpList = new WX_LISTBOX( fpPanel, ID_MODVIEW_FOOTPRINT_LIST, wxDefaultPosition, wxDefaultSize,
+                               0, nullptr, wxLB_HSCROLL | wxNO_BORDER );
+
+    m_fpList->Connect( wxEVT_LEFT_DCLICK,
+                       wxMouseEventHandler( FOOTPRINT_VIEWER_FRAME::DClickOnFootprintList ),
+                       nullptr, this );
     fpSizer->Add( m_fpList, 1, wxEXPAND, 5 );
 
     fpPanel->SetSizer( fpSizer );
@@ -192,11 +193,11 @@ FOOTPRINT_VIEWER_FRAME::FOOTPRINT_VIEWER_FRAME( KIWAY* aKiway, wxWindow* aParent
     // In viewer, the default net clearance is not known (it depends on the actual board).
     // So we do not show the default clearance, by setting it to 0
     // The footprint or pad specific clearance will be shown
-    GetBoard()->GetDesignSettings().GetDefault()->SetClearance( 0 );
+    GetBoard()->GetDesignSettings().m_NetSettings->GetDefaultNetclass()->SetClearance( 0 );
 
     // Don't show the default board solder mask clearance in the footprint viewer.  Only the
     // footprint or pad clearance setting should be shown if it is not 0.
-    GetBoard()->GetDesignSettings().m_SolderMaskMargin = 0;
+    GetBoard()->GetDesignSettings().m_SolderMaskExpansion = 0;
 
     // Ensure all layers and items are visible:
     GetBoard()->SetVisibleAlls();
@@ -206,9 +207,6 @@ FOOTPRINT_VIEWER_FRAME::FOOTPRINT_VIEWER_FRAME( KIWAY* aKiway, wxWindow* aParent
     LoadSettings( config() );
     GetGalDisplayOptions().m_axesEnabled = true;
 
-    // Call resolveCanvasType after loading settings:
-    resolveCanvasType();
-
     // Create the manager and dispatcher & route draw panel events to the dispatcher
     m_toolManager = new TOOL_MANAGER;
     m_toolManager->SetEnvironment( GetBoard(), drawPanel->GetView(),
@@ -239,6 +237,9 @@ FOOTPRINT_VIEWER_FRAME::FOOTPRINT_VIEWER_FRAME( KIWAY* aKiway, wxWindow* aParent
     ReCreateLibraryList();
     UpdateTitle();
 
+    // Call resolveCanvasType after loading settings:
+    resolveCanvasType();
+
     // If a footprint was previously loaded, reload it
     if( getCurNickname().size() && getCurFootprintName().size() )
     {
@@ -246,7 +247,14 @@ FOOTPRINT_VIEWER_FRAME::FOOTPRINT_VIEWER_FRAME( KIWAY* aKiway, wxWindow* aParent
 
         id.SetLibNickname( getCurNickname() );
         id.SetLibItemName( getCurFootprintName() );
-        GetBoard()->Add( loadFootprint( id ) );
+
+        FOOTPRINT* footprint = loadFootprint( id );
+
+        if( footprint )
+        {
+            GetBoard()->Add( footprint );
+            setFPWatcher( footprint );
+        }
     }
 
     drawPanel->DisplayBoard( m_pcb );
@@ -261,7 +269,7 @@ FOOTPRINT_VIEWER_FRAME::FOOTPRINT_VIEWER_FRAME( KIWAY* aKiway, wxWindow* aParent
     // Vertical items; layers 1 - 3
     m_auimgr.AddPane( libPanel, EDA_PANE().Palette().Name( "Libraries" ).Left().Layer(2)
                       .CaptionVisible( false ).MinSize( 100, -1 ).BestSize( 200, -1 ) );
-    m_auimgr.AddPane( fpPanel, EDA_PANE().Palette().Name( "Footprints" ).Left().Layer( 1)
+    m_auimgr.AddPane( fpPanel, EDA_PANE().Palette().Name( "Footprints" ).Left().Layer(1)
                       .CaptionVisible( false ).MinSize( 100, -1 ).BestSize( 300, -1 ) );
 
     m_auimgr.AddPane( GetCanvas(), EDA_PANE().Canvas().Name( "DrawFrame" ).Center() );
@@ -269,25 +277,35 @@ FOOTPRINT_VIEWER_FRAME::FOOTPRINT_VIEWER_FRAME( KIWAY* aKiway, wxWindow* aParent
     // after changing something to the aui manager call Update() to reflect the changes
     m_auimgr.Update();
 
+    if( m_libListWidth > 0 )
+        SetAuiPaneSize( m_auimgr, m_auimgr.GetPane( "Libraries" ), m_libListWidth, -1 );
+
+    if( m_fpListWidth > 0 )
+        SetAuiPaneSize( m_auimgr, m_auimgr.GetPane( "Footprints" ), m_fpListWidth, -1 );
+
     // The canvas should not steal the focus from the list boxes
     GetCanvas()->SetCanFocus( false );
     GetCanvas()->GetGAL()->SetAxesEnabled( true );
     ActivateGalCanvas();
 
-    // Restore last zoom.  (If auto-zooming we'll adjust when we load the footprint.)
+    // Restore last zoom and auto zoom option.  (If auto-zooming we'll adjust when we load the footprint.)
     PCBNEW_SETTINGS* cfg = GetPcbNewSettings();
     wxASSERT( cfg );
     GetCanvas()->GetView()->SetScale( cfg->m_FootprintViewerZoom );
 
+    wxAuiToolBarItem* toolOpt = m_mainToolBar->FindTool( ID_FPVIEWER_AUTOZOOM_TOOL );
+
+    if( cfg->m_FootprintViewerAutoZoomOnSelect )
+        toolOpt->SetState( wxAUI_BUTTON_STATE_CHECKED );
+    else
+        toolOpt->SetState( 0 );
+
     updateView();
     setupUnits( config() );
 
-    if( !IsModal() )        // For modal mode, calling ShowModal() will show this frame
-    {
-        ReCreateFootprintList();
-        Raise();            // On some window managers, this is needed
-        Show( true );
-    }
+    ReCreateFootprintList();
+    Raise();            // On some window managers, this is needed
+    Show( true );
 }
 
 
@@ -301,6 +319,10 @@ FOOTPRINT_VIEWER_FRAME::~FOOTPRINT_VIEWER_FRAME()
     GetCanvas()->GetView()->Clear();
     // Be sure any event cannot be fired after frame deletion:
     GetCanvas()->SetEvtHandlerEnabled( false );
+    m_fpList->Disconnect( wxEVT_LEFT_DCLICK,
+                          wxMouseEventHandler( FOOTPRINT_VIEWER_FRAME::DClickOnFootprintList ),
+                          nullptr, this );
+    setFPWatcher( nullptr );
 }
 
 
@@ -310,6 +332,19 @@ SELECTION& FOOTPRINT_VIEWER_FRAME::GetCurrentSelection()
 }
 
 
+void FOOTPRINT_VIEWER_FRAME::UpdateMsgPanel()
+{
+    EDA_DRAW_FRAME::UpdateMsgPanel();
+
+    if( FOOTPRINT* fp = static_cast<FOOTPRINT*>( GetModel() ) )
+    {
+        std::vector<MSG_PANEL_ITEM> msgItems;
+        fp->GetMsgPanelInfo( this, msgItems );
+        SetMsgPanel( msgItems );
+    }
+}
+
+
 void FOOTPRINT_VIEWER_FRAME::setupUIConditions()
 {
     PCB_BASE_FRAME::setupUIConditions();
@@ -322,12 +357,11 @@ void FOOTPRINT_VIEWER_FRAME::setupUIConditions()
 #define ENABLE( x ) ACTION_CONDITIONS().Enable( x )
 #define CHECK( x )  ACTION_CONDITIONS().Check( x )
 
-    mgr->SetConditions( ACTIONS::toggleGrid,             CHECK( cond.GridVisible() ) );
-    mgr->SetConditions( ACTIONS::toggleCursorStyle,      CHECK( cond.FullscreenCursor() ) );
-    mgr->SetConditions( ACTIONS::millimetersUnits,
-                        CHECK( cond.Units( EDA_UNITS::MILLIMETRES ) ) );
-    mgr->SetConditions( ACTIONS::inchesUnits,            CHECK( cond.Units( EDA_UNITS::INCHES ) ) );
-    mgr->SetConditions( ACTIONS::milsUnits,              CHECK( cond.Units( EDA_UNITS::MILS ) ) );
+    mgr->SetConditions( ACTIONS::toggleGrid,        CHECK( cond.GridVisible() ) );
+    mgr->SetConditions( ACTIONS::toggleCursorStyle, CHECK( cond.FullscreenCursor() ) );
+    mgr->SetConditions( ACTIONS::millimetersUnits,  CHECK( cond.Units( EDA_UNITS::MM ) ) );
+    mgr->SetConditions( ACTIONS::inchesUnits,       CHECK( cond.Units( EDA_UNITS::INCH ) ) );
+    mgr->SetConditions( ACTIONS::milsUnits,         CHECK( cond.Units( EDA_UNITS::MILS ) ) );
 
 
     mgr->SetConditions( ACTIONS::zoomTool,
@@ -341,6 +375,7 @@ void FOOTPRINT_VIEWER_FRAME::setupUIConditions()
     mgr->SetConditions( PCB_ACTIONS::padDisplayMode,     CHECK( !cond.PadFillDisplay() ) );
     mgr->SetConditions( PCB_ACTIONS::textOutlines,       CHECK( !cond.TextFillDisplay() ) );
     mgr->SetConditions( PCB_ACTIONS::graphicsOutlines,   CHECK( !cond.GraphicsFillDisplay() ) );
+    mgr->SetConditions( ACTIONS::toggleBoundingBoxes,    CHECK( cond.BoundingBoxes() ) );
 
 #undef ENABLE
 #undef CHECK
@@ -356,19 +391,7 @@ void FOOTPRINT_VIEWER_FRAME::doCloseWindow()
 
     GetCanvas()->StopDrawing();
 
-    if( IsModal() )
-    {
-        // Only dismiss a modal frame once, so that the return values set by
-        // the prior DismissModal() are not bashed for ShowModal().
-        if( !IsDismissed() )
-            DismissModal( false );
-
-        // window to be destroyed by the caller of KIWAY_PLAYER::ShowModal()
-    }
-    else
-    {
-        Destroy();
-    }
+    Destroy();
 }
 
 
@@ -385,32 +408,53 @@ void FOOTPRINT_VIEWER_FRAME::ReCreateLibraryList()
 {
     m_libList->Clear();
 
-    std::vector<wxString> nicknames = Prj().PcbFootprintLibs()->GetLogicalLibs();
-    std::set<wxString>    excludes;
+    COMMON_SETTINGS*      cfg = Pgm().GetCommonSettings();
+    PROJECT_FILE&         project = Kiway().Prj().GetProjectFile();
+    std::vector<wxString> nicknames = PROJECT_PCB::PcbFootprintLibs( &Prj() )->GetLogicalLibs();
+    std::vector<wxString> pinnedMatches;
+    std::vector<wxString> otherMatches;
 
-    if( !m_libFilter->GetValue().IsEmpty() )
+    auto process =
+            [&]( const wxString& aNickname )
+            {
+                if( alg::contains( project.m_PinnedFootprintLibs, aNickname )
+                        || alg::contains( cfg->m_Session.pinned_fp_libs, aNickname ) )
+                {
+                    pinnedMatches.push_back( aNickname );
+                }
+                else
+                {
+                    otherMatches.push_back( aNickname );
+                }
+            };
+
+    if( m_libFilter->GetValue().IsEmpty() )
+    {
+        for( const wxString& nickname : nicknames )
+            process( nickname );
+    }
+    else
     {
         wxStringTokenizer tokenizer( m_libFilter->GetValue() );
 
         while( tokenizer.HasMoreTokens() )
         {
             const wxString       term = tokenizer.GetNextToken().Lower();
-            EDA_COMBINED_MATCHER matcher( term );
-            int                  matches, position;
+            EDA_COMBINED_MATCHER matcher( term, CTX_LIBITEM );
 
             for( const wxString& nickname : nicknames )
             {
-                if( !matcher.Find( nickname.Lower(), matches, position ) )
-                    excludes.insert( nickname );
+                if( matcher.Find( nickname.Lower() ) )
+                    process( nickname );
             }
         }
     }
 
-    for( const wxString& nickname : nicknames )
-    {
-        if( !excludes.count( nickname ) )
-            m_libList->Append( nickname );
-    }
+    for( const wxString& nickname : pinnedMatches )
+        m_libList->Append( LIB_TREE_MODEL_ADAPTER::GetPinningSymbol() + nickname );
+
+    for( const wxString& nickname : otherMatches )
+        m_libList->Append( nickname );
 
     // Search for a previous selection:
     int index =  m_libList->FindString( getCurNickname(), true );
@@ -451,7 +495,7 @@ void FOOTPRINT_VIEWER_FRAME::ReCreateFootprintList()
 
     wxString nickname = getCurNickname();
 
-    fp_info_list->ReadFootprintFiles( Prj().PcbFootprintLibs(), !nickname ? nullptr : &nickname );
+    fp_info_list->ReadFootprintFiles( PROJECT_PCB::PcbFootprintLibs( &Prj() ), !nickname ? nullptr : &nickname );
 
     if( fp_info_list->GetErrorCount() )
     {
@@ -471,17 +515,16 @@ void FOOTPRINT_VIEWER_FRAME::ReCreateFootprintList()
 
         while( tokenizer.HasMoreTokens() )
         {
-            const wxString       term = tokenizer.GetNextToken().Lower();
-            EDA_COMBINED_MATCHER matcher( term );
-            int                  matches, position;
+            const wxString       filterTerm = tokenizer.GetNextToken().Lower();
+            EDA_COMBINED_MATCHER matcher( filterTerm, CTX_LIBITEM );
 
             for( const std::unique_ptr<FOOTPRINT_INFO>& footprint : fp_info_list->GetList() )
             {
-                wxString search = footprint->GetFootprintName() + wxS( " " ) + footprint->GetSearchText();
-                bool     matched = matcher.Find( search.Lower(), matches, position );
+                std::vector<SEARCH_TERM> searchTerms = footprint->GetSearchTerms();
+                int                      matched = matcher.ScoreTerms( searchTerms );
 
-                if( !matched && term.IsNumber() )
-                    matched = ( wxAtoi( term ) == (int)footprint->GetPadCount() );
+                if( filterTerm.IsNumber() && wxAtoi( filterTerm ) == (int)footprint->GetPadCount() )
+                    matched++;
 
                 if( !matched )
                     excludes.insert( footprint->GetFootprintName() );
@@ -570,7 +613,8 @@ void FOOTPRINT_VIEWER_FRAME::OnCharHook( wxKeyEvent& aEvent )
         else
             aEvent.Skip();
     }
-    else if( aEvent.GetKeyCode() == WXK_RETURN && m_fpList->GetSelection() >= 0 )
+    else if( ( aEvent.GetKeyCode() == WXK_RETURN || aEvent.GetKeyCode() == WXK_NUMPAD_ENTER )
+             && m_fpList->GetSelection() >= 0 )
     {
         wxCommandEvent dummy;
         AddFootprintToPCB( dummy );
@@ -582,7 +626,7 @@ void FOOTPRINT_VIEWER_FRAME::OnCharHook( wxKeyEvent& aEvent )
 }
 
 
-void FOOTPRINT_VIEWER_FRAME::selectPrev( wxListBox* aListBox )
+void FOOTPRINT_VIEWER_FRAME::selectPrev( WX_LISTBOX* aListBox )
 {
     int prev = aListBox->GetSelection() - 1;
 
@@ -601,7 +645,7 @@ void FOOTPRINT_VIEWER_FRAME::selectPrev( wxListBox* aListBox )
 }
 
 
-void FOOTPRINT_VIEWER_FRAME::selectNext( wxListBox* aListBox )
+void FOOTPRINT_VIEWER_FRAME::selectNext( WX_LISTBOX* aListBox )
 {
     int next = aListBox->GetSelection() + 1;
 
@@ -627,13 +671,16 @@ void FOOTPRINT_VIEWER_FRAME::ClickOnLibList( wxCommandEvent& aEvent )
     if( ii < 0 )
         return;
 
-    wxString name = m_libList->GetString( ii );
+    wxString name = m_libList->GetBaseString( ii );
 
     if( getCurNickname() == name )
         return;
 
     setCurNickname( name );
 
+    // Ensure the displayed footprint is loade/reloaded from the new library
+    setCurFootprintName( wxEmptyString );
+
     ReCreateFootprintList();
     UpdateTitle();
 }
@@ -649,67 +696,48 @@ void FOOTPRINT_VIEWER_FRAME::ClickOnFootprintList( wxCommandEvent& aEvent )
     if( ii < 0 )
         return;
 
-    wxString name = m_fpList->GetString( ii );
+    wxString name = m_fpList->GetBaseString( ii );
 
     if( getCurFootprintName().CmpNoCase( name ) != 0 )
     {
         setCurFootprintName( name );
-
-        // Delete the current footprint (MUST reset tools first)
-        GetToolManager()->ResetTools( TOOL_BASE::MODEL_RELOAD );
-
-        GetBoard()->DeleteAllFootprints();
-
-        LIB_ID id;
-        id.SetLibNickname( getCurNickname() );
-        id.SetLibItemName( getCurFootprintName() );
-
-        try
-        {
-            GetBoard()->Add( loadFootprint( id ) );
-        }
-        catch( const IO_ERROR& ioe )
-        {
-            wxString msg =
-                    wxString::Format( _( "Could not load footprint '%s' from library '%s'."
-                                         "\n\n%s" ),
-                                      getCurFootprintName(), getCurNickname(), ioe.Problem() );
-            DisplayError( this, msg );
-        }
-
-        UpdateTitle();
-
-        updateView();
-
-        GetCanvas()->Refresh();
-        Update3DView( true, true );
+        SelectAndViewFootprint( NEW_PART );
     }
 }
 
 
-void FOOTPRINT_VIEWER_FRAME::DClickOnFootprintList( wxCommandEvent& aEvent )
+void FOOTPRINT_VIEWER_FRAME::displayFootprint( FOOTPRINT* aFootprint )
 {
-    AddFootprintToPCB( aEvent );
+    for( PAD* pad : aFootprint->Pads() )
+    {
+        const COMPONENT_NET& net = m_comp.GetNet( pad->GetNumber() );
+
+        if( !net.GetPinFunction().IsEmpty() )
+        {
+            NETINFO_ITEM* netinfo = new NETINFO_ITEM( GetBoard() );
+            netinfo->SetNetname( net.GetPinFunction() );
+            GetBoard()->Add( netinfo );
+            pad->SetNet( netinfo );
+        }
+    }
+
+    GetBoard()->Add( aFootprint );
+}
+
+
+void FOOTPRINT_VIEWER_FRAME::DClickOnFootprintList( wxMouseEvent& aEvent )
+{
+    wxCommandEvent evt;
+    AddFootprintToPCB( evt );
 }
 
 
 void FOOTPRINT_VIEWER_FRAME::AddFootprintToPCB( wxCommandEvent& aEvent )
 {
-    if( IsModal() )
+    if( GetBoard()->GetFirstFootprint() )
     {
-        if( m_fpList->GetSelection() >= 0 )
-        {
-            LIB_ID fpid( getCurNickname(), m_fpList->GetStringSelection() );
-            DismissModal( true, fpid.Format() );
-        }
-        else
-        {
-            DismissModal( false );
-        }
-    }
-    else if( GetBoard()->GetFirstFootprint() )
-    {
-        PCB_EDIT_FRAME* pcbframe = (PCB_EDIT_FRAME*) Kiway().Player( FRAME_PCB_EDITOR, false );
+        PCB_EDIT_FRAME*  pcbframe = (PCB_EDIT_FRAME*) Kiway().Player( FRAME_PCB_EDITOR, false );
+        PCBNEW_SETTINGS* cfg = pcbframe->GetPcbNewSettings();
 
         if( pcbframe == nullptr )      // happens when the board editor is not active (or closed)
         {
@@ -725,7 +753,12 @@ void FOOTPRINT_VIEWER_FRAME::AddFootprintToPCB( wxCommandEvent& aEvent )
             return;
         }
 
-        toolMgr->RunAction( PCB_ACTIONS::selectionClear, true );
+        wxWindow* blocking_dialog = pcbframe->Kiway().GetBlockingDialog();
+
+        if( blocking_dialog )
+            blocking_dialog->Close( true );
+
+        toolMgr->RunAction( PCB_ACTIONS::selectionClear );
         BOARD_COMMIT commit( pcbframe );
 
         // Create the "new" footprint
@@ -737,7 +770,7 @@ void FOOTPRINT_VIEWER_FRAME::AddFootprintToPCB( wxCommandEvent& aEvent )
         for( PAD* pad : newFootprint->Pads() )
         {
             // Set the pads ratsnest settings to the global settings
-            pad->SetLocalRatsnestVisible( pcbframe->GetDisplayOptions().m_ShowGlobalRatsnest );
+            pad->SetLocalRatsnestVisible( cfg->m_Display.m_ShowGlobalRatsnest );
 
             // Pads in the library all have orphaned nets.  Replace with Default.
             pad->SetNetCode( 0 );
@@ -746,8 +779,7 @@ void FOOTPRINT_VIEWER_FRAME::AddFootprintToPCB( wxCommandEvent& aEvent )
         // Put it on FRONT layer,
         // (Can be stored flipped if the lib is an archive built from a board)
         if( newFootprint->IsFlipped() )
-            newFootprint->Flip( newFootprint->GetPosition(),
-                                pcbframe->Settings().m_FlipLeftRight );
+            newFootprint->Flip( newFootprint->GetPosition(), cfg->m_FlipDirection );
 
         KIGFX::VIEW_CONTROLS* viewControls = pcbframe->GetCanvas()->GetViewControls();
         VECTOR2D              cursorPos = viewControls->GetCursorPosition();
@@ -756,12 +788,12 @@ void FOOTPRINT_VIEWER_FRAME::AddFootprintToPCB( wxCommandEvent& aEvent )
         viewControls->SetCrossHairCursorPosition( VECTOR2D( 0, 0 ), false );
         pcbframe->PlaceFootprint( newFootprint );
 
-        newFootprint->SetPosition( wxPoint( 0, 0 ) );
+        newFootprint->SetPosition( VECTOR2I( 0, 0 ) );
         viewControls->SetCrossHairCursorPosition( cursorPos, false );
-        commit.Push( wxT( "Insert footprint" ) );
+        commit.Push( _( "Insert Footprint" ) );
 
         pcbframe->Raise();
-        toolMgr->RunAction( PCB_ACTIONS::placeFootprint, true, newFootprint );
+        toolMgr->PostAction( PCB_ACTIONS::placeFootprint, newFootprint );
 
         newFootprint->ClearFlags();
     }
@@ -779,9 +811,23 @@ void FOOTPRINT_VIEWER_FRAME::LoadSettings( APP_SETTINGS_BASE* aCfg )
     PCB_BASE_FRAME::LoadSettings( aCfg );
 
     // Fetch display and grid settings from Footprint Editor
-    auto fpedit = Pgm().GetSettingsManager().GetAppSettings<FOOTPRINT_EDITOR_SETTINGS>();
+    SETTINGS_MANAGER&          mgr = Pgm().GetSettingsManager();
+    FOOTPRINT_EDITOR_SETTINGS* fpedit = mgr.GetAppSettings<FOOTPRINT_EDITOR_SETTINGS>( "fpedit" );
+
     m_displayOptions = fpedit->m_Display;
     GetGalDisplayOptions().ReadWindowSettings( fpedit->m_Window );
+
+    m_libListWidth = cfg->m_FootprintViewerLibListWidth;
+    m_fpListWidth = cfg->m_FootprintViewerFPListWidth;
+
+    // Set parameters to a reasonable value.
+    int maxWidth = cfg->m_FootprintViewer.state.size_x - 80;
+
+    if( m_libListWidth + m_fpListWidth > maxWidth )
+    {
+        m_libListWidth = maxWidth * ( m_libListWidth / ( m_libListWidth + m_fpListWidth ) );
+        m_fpListWidth = maxWidth - m_libListWidth;
+    }
 }
 
 
@@ -797,6 +843,12 @@ void FOOTPRINT_VIEWER_FRAME::SaveSettings( APP_SETTINGS_BASE* aCfg )
 
     if( GetCanvas() && GetCanvas()->GetView() )
         cfg->m_FootprintViewerZoom = GetCanvas()->GetView()->GetScale();
+
+    wxAuiToolBarItem* toolOpt = m_mainToolBar->FindTool( ID_FPVIEWER_AUTOZOOM_TOOL );
+    cfg->m_FootprintViewerAutoZoomOnSelect = ( toolOpt->GetState() & wxAUI_BUTTON_STATE_CHECKED );
+    cfg->m_FootprintViewerLibListWidth = m_libList->GetSize().x;
+    cfg->m_FootprintViewerFPListWidth = m_fpList->GetSize().x;
+
 }
 
 
@@ -809,22 +861,25 @@ WINDOW_SETTINGS* FOOTPRINT_VIEWER_FRAME::GetWindowSettings( APP_SETTINGS_BASE* a
 }
 
 
-COLOR_SETTINGS* FOOTPRINT_VIEWER_FRAME::GetColorSettings() const
+COLOR_SETTINGS* FOOTPRINT_VIEWER_FRAME::GetColorSettings( bool aForceRefresh ) const
 {
-    auto* settings = Pgm().GetSettingsManager().GetAppSettings<FOOTPRINT_EDITOR_SETTINGS>();
+    SETTINGS_MANAGER&          mgr = Pgm().GetSettingsManager();
+    FOOTPRINT_EDITOR_SETTINGS* cfg = mgr.GetAppSettings<FOOTPRINT_EDITOR_SETTINGS>( "fpedit" );
 
-    if( settings )
-        return Pgm().GetSettingsManager().GetColorSettings( settings->m_ColorTheme );
+    if( cfg )
+        return mgr.GetColorSettings( cfg->m_ColorTheme );
     else
-        return Pgm().GetSettingsManager().GetColorSettings();
+        return mgr.GetColorSettings();
 }
 
 
-void FOOTPRINT_VIEWER_FRAME::CommonSettingsChanged( bool aEnvVarsChanged, bool aTextVarsChanged )
+void FOOTPRINT_VIEWER_FRAME::CommonSettingsChanged( int aFlags )
 {
-    PCB_BASE_FRAME::CommonSettingsChanged( aEnvVarsChanged, aTextVarsChanged );
+    PCB_BASE_FRAME::CommonSettingsChanged( aFlags );
 
-    if( aEnvVarsChanged )
+    GetCanvas()->ForceRefresh();
+
+    if( aFlags & ENVVARS_CHANGED )
         ReCreateLibraryList();
 }
 
@@ -858,7 +913,7 @@ void FOOTPRINT_VIEWER_FRAME::OnActivate( wxActivateEvent& event )
     if( event.GetActive() )
     {
         // Ensure we have the right library list:
-        std::vector< wxString > libNicknames = Prj().PcbFootprintLibs()->GetLogicalLibs();
+        std::vector< wxString > libNicknames = PROJECT_PCB::PcbFootprintLibs( &Prj() )->GetLogicalLibs();
         bool                    stale = false;
 
         if( libNicknames.size() != m_libList->GetCount() )
@@ -867,7 +922,7 @@ void FOOTPRINT_VIEWER_FRAME::OnActivate( wxActivateEvent& event )
         {
             for( unsigned ii = 0;  ii < libNicknames.size();  ii++ )
             {
-                if( libNicknames[ii] != m_libList->GetString( ii ) )
+                if( libNicknames[ii] != m_libList->GetBaseString( ii ) )
                 {
                     stale = true;
                     break;
@@ -892,53 +947,32 @@ void FOOTPRINT_VIEWER_FRAME::OnUpdateFootprintButton( wxUpdateUIEvent& aEvent )
 }
 
 
-bool FOOTPRINT_VIEWER_FRAME::ShowModal( wxString* aFootprint, wxWindow* aParent )
+void FOOTPRINT_VIEWER_FRAME::ReloadFootprint( FOOTPRINT* aFootprint )
 {
-    if( aFootprint && !aFootprint->IsEmpty() )
-    {
-        wxString msg;
-        LIB_TABLE* fpTable = Prj().PcbFootprintLibs();
-        LIB_ID fpid;
+    setCurNickname( aFootprint->GetFPID().GetLibNickname() );
+    setCurFootprintName( aFootprint->GetFPID().GetLibItemName() );
+    SelectAndViewFootprint( RELOAD_PART );
+}
 
-        fpid.Parse( *aFootprint, true );
 
-        if( fpid.IsValid() )
-        {
-            wxString libraryName = fpid.GetLibNickname();
-
-            if( !fpTable->HasLibrary( fpid.GetLibNickname(), false ) )
-            {
-                msg.sprintf( _( "The current configuration does not include library '%s'.  Use "
-                                "Manage Footprint Libraries to edit the configuration." ),
-                             libraryName );
-                DisplayErrorMessage( aParent, _( "Footprint library not found." ), msg );
-            }
-            else if ( !fpTable->HasLibrary( fpid.GetLibNickname(), true ) )
-            {
-                msg.sprintf( _( "Library '%s' is not enabled in the current configuration.  Use "
-                                "Manage Footprint Libraries to edit the configuration." ),
-                             libraryName );
-                DisplayErrorMessage( aParent, _( "Footprint library not enabled." ), msg );
-            }
-            else
-            {
-                // Update last selection:
-                setCurNickname( libraryName );
-                setCurFootprintName( fpid.GetLibItemName() );
-                m_libList->SetStringSelection( libraryName );
-            }
-        }
-    }
-
-    // Rebuild the fp list from the last selected library,
-    // and show the last selected footprint
+void FOOTPRINT_VIEWER_FRAME::HardRedraw()
+{
+    ReCreateLibraryList();
     ReCreateFootprintList();
-    SelectAndViewFootprint( NEW_PART );
+    ReloadFootprint( GetBoard()->GetFirstFootprint() );
+}
 
-    bool retval = KIWAY_PLAYER::ShowModal( aFootprint, aParent );
+void FOOTPRINT_VIEWER_FRAME::KiwayMailIn( KIWAY_EXPRESS& mail )
+{
+    switch( mail.Command() )
+    {
+    case MAIL_RELOAD_LIB:
+        ReCreateLibraryList();
+        break;
 
-    m_libFilter->SetFocus();
-    return retval;
+    default:
+        break;
+    }
 }
 
 
@@ -977,17 +1011,20 @@ void FOOTPRINT_VIEWER_FRAME::OnIterateFootprintList( wxCommandEvent& event )
 void FOOTPRINT_VIEWER_FRAME::UpdateTitle()
 {
     wxString title;
-    wxString path;
 
     if( !getCurNickname().IsEmpty() )
     {
-        title = getCurNickname();
+        try
+        {
+            FP_LIB_TABLE* libtable = PROJECT_PCB::PcbFootprintLibs( &Prj() );
+            const LIB_TABLE_ROW* row = libtable->FindRow( getCurNickname() );
 
-        FP_LIB_TABLE* libtable = Prj().PcbFootprintLibs();
-        const LIB_TABLE_ROW* row = libtable->FindRow( getCurNickname() );
-
-        if( row )
-            title += wxT( " \u2014 " ) + row->GetFullURI( true );
+            title = getCurNickname() + wxT( " \u2014 " ) + row->GetFullURI( true );
+        }
+        catch( ... )
+        {
+            title = _( "[no library selected]" );
+        }
     }
     else
     {
@@ -1024,19 +1061,25 @@ void FOOTPRINT_VIEWER_FRAME::SelectAndViewFootprint( int aMode )
         m_fpList->SetSelection( selection );
         m_fpList->EnsureVisible( selection );
 
-        setCurFootprintName( m_fpList->GetString((unsigned) selection ) );
+        setCurFootprintName( m_fpList->GetBaseString( selection ) );
 
         // Delete the current footprint
-        GetBoard()->DeleteAllFootprints();
+        if( PCB_SELECTION_TOOL* selectionTool = m_toolManager->GetTool<PCB_SELECTION_TOOL>() )
+            selectionTool->ClearSelection( true /* quiet mode */ );
 
-        FOOTPRINT* footprint = Prj().PcbFootprintLibs()->FootprintLoad( getCurNickname(),
+        GetBoard()->DeleteAllFootprints();
+        GetBoard()->RemoveUnusedNets( nullptr );
+
+        FOOTPRINT* footprint = PROJECT_PCB::PcbFootprintLibs( &Prj() )->FootprintLoad( getCurNickname(),
                                                                         getCurFootprintName() );
 
         if( footprint )
-            GetBoard()->Add( footprint, ADD_MODE::APPEND );
+            displayFootprint( footprint );
+
+        if( aMode != RELOAD_PART )
+            setFPWatcher( footprint );
 
         Update3DView( true, true );
-
         updateView();
     }
 
@@ -1053,10 +1096,12 @@ void FOOTPRINT_VIEWER_FRAME::updateView()
 
     m_toolManager->ResetTools( TOOL_BASE::MODEL_RELOAD );
 
-    if( m_zoomSelectBox->GetSelection() == 0 )
-        m_toolManager->RunAction( ACTIONS::zoomFitScreen, true );
+    wxAuiToolBarItem* toolOpt = m_mainToolBar->FindTool( ID_FPVIEWER_AUTOZOOM_TOOL );
+
+    if( toolOpt->GetState() & wxAUI_BUTTON_STATE_CHECKED )
+        m_toolManager->RunAction( ACTIONS::zoomFitScreen );
     else
-        m_toolManager->RunAction( ACTIONS::centerContents, true );
+        m_toolManager->RunAction( ACTIONS::centerContents );
 
     UpdateMsgPanel();
 }
